
unit    Marquee;

{   MARQUEE.PAS - support for an "LCD style" marquee that contains up to forty
    characters across that can be drawn in cool ways. Functions as either a
    label or a progress meter.

    Copyright (C) 1997 Midnight Software, Inc., All Rights Reserved.

    Installation:

    Delphi 1.0: Copy MARQUEE.PAS and MARQUEE.D16 into your components library
    directory. Rename MARQUEE.D16 to MARQUEE.DCR. Add MARQUEE.PAS to your
    library.

    Delphi 2.0 and above: Copy MARQUEE.PAS and MARQUEE.DCR to your components
    library directory. Add MARQUEE.PAS to your library.

    Use:

    Place the component onto your form (d-uh???)

    Always set these properties:
    BrightColor         the colour of the "lit" pixels
    DimColor            the colour of the "dark" pixels (usu. Navy)
    PipSize             the size of a "pixel". 

    If you want to use the Marquee as a progress indicator, set these:
    ShowProgressPercent TRUE = print "XX%" in the middle of the bar,
                        FALSE = just show the meter running across.
    Progress            n/ProgressMax - current completion percentage
    ProgressMax         Progress value indicating 100% completion

    If you want to use the Marquee as a label, set these:
    MaxChars            Leave alone to fill the cell with characters, or
                        specify a >0 value to limit to that many characters.

    Caption             Just like any other caption

    CodePage            leave as 0 for default codepage, otherwise set to
                        1250 (Eastern Europe) or 1252 (Western Europe)
                        NOTE: Delphi 1.0 users: only supported "default"
                        codepage is 1252 -- Western Europe. 

    AnimationSpeed      anFast      as fast as it will draw
                        anMedium    not quite as fast
                        anSlow      slow

    AnimationStyle      dmNormal    just paint the text

                        dmSlide, dmVerticalSlide, dmSlideChase,   
                        dmWipe, dmNewWhite, dmRunLeft, dmRunRight,     
                        dmChaseLeft, dmChaseRight, dmPopUp, dmPaintRandom,  
                        dmPopUpNew,  dmPopDownNew,   
                        dmPopDown:      various drawing effects - try 'em all!


    FREEWARE NOTICE: You may incorporate this component into any derivitave
    work including component libraries and executable programs, with no
    consideration to Midnight Software whatsoever, except as noted below.

    You may NOT include this component in any component library for which
    there is a charge of greater than USA$50.00, UNLESS the purpose of this
    library is primary educational (e.g., as part of a Delphi book or
    companion CD-ROM)

    If you are going to redistribute this component alone or with other
    components as a component library, you MUST include all source code.

    This component may NOT be modified to function as an ActiveX(tm) component,
    nor may it be modified in any other way to make it compatible with any
    Microsoft(tm) software development product. Midnight Software's intent
    is that this component is for the exclusive use of developers using
    Borland tools. 

    ActiveX(tm) and Microsoft(tm) are registered trademarks of Microsoft
    Corporation.


    ... and PLEASE ...

    * Let me know who you are!!! If you are using this component, please
    let me know if it suits your needs! My e-mail address is 
    GutterBoy@mindless.com

    * If you make any significant enhancements or correct any bugs, I would
    appreciate a copy.


    ... and, of course ...

    Midnight Software makes no claims about the merchantability of this
    product. THIS COMPONENT IS PROVIDED "AS IS" WITH NO WARRANTIES NOR
    GUARANTEES OF ANY KIND. By using this component, you agree to assume all 
    risks associated with its use or misuse. Midnight Software, Inc., shall 
    be held harmless for any incidental or consequential damages resulting 
    from the use or misuse of this component.

    REVISION HISTORY

    31 Dec 1996 rdk Added AnimationSpeed property

    *****       VERSION 1.00        *****

    11 Jul 1997 rdk Replaced MoveTo/LineTo with Rectangle or Pixel. It's a _lot_ faster!

    *****       VERSION 1.01        *****

    24 Jul 1997 rdk Support for Eastern/Western European Code Pages.
                    Thanks to Ivo Eichler for the idea and the bitmaps!
                    dmDrawNewWhite and dmPaintRandom now work!

    END OF HISTORY

    To do list:
    * Add other effects: OpenDoors - new text replaces old from the middle two chars
    outward. CloseDoors - reverse of opendoors.
    * implement mechanicalEffect flag -- before text stops, it goes one pixel too far then
    backs up into place.
    * support for a display of several rows of characters.
    * support for 3-D pips when pixel size is 3 or greater:
        FPixelAppearance = paFlat, paInset, paRaised

        (L=light color, N=normal pip color, S=shadow pip color)
        -   special effect that has pips appear to "rise" and "fall" out of
            and into the screen.

    3:      4:      5:

    LLN     LLLN    LLLLN
    LNS     LNNS    LNNNS
    NSS     LNNS    LNNNS
            NSSS    LNNNS
                    NSSSS
}

interface

uses    SysUtils, Classes, Graphics, ExtCtrls;

const
    PIXEL_HEIGHT    = 9;
    PROGRESS_METER  = -1;
    MAXCHARS_AUTO   = 0;        { allow frame dimensions to dictate max width }
    MAX_CHARS       = 40;       { max. characters permitted irregardless }
    MAXWIDTH        = ( 6 * MAX_CHARS ) - 1;    { number of pips across the screen }

type
    String32 = String[ MAX_CHARS ];

type
{   an AnimationSpeed of anSlow pauses 3/18 second between draws
    an AnimationSpeed of anMedium pauses 1/18 second between draws
    an AnimationSPeed of anFast does not pause between draws.
}

    TPixelAppearance = ( paFlat, paInset, paRaised );
    TAnimationSpeed = ( anSlow, anMedium, anFast );

{   Following are the possible draw modes. Modes with descriptions preceded by
    'M' are affected by the setting of mechanicalEffect,
    'A' are affected by the setting of animationSpeed
}
    TDrawMode = (
                    dmRandom,       {    select a mode at random }
                    dmNormal,       {    just display the information - default }
                    dmSlide,        {  A alt. rows slide out r/l, then slide in }
                    dmVerticalSlide,{  A alt. cols slide out u/d, then slide in }
                    dmSlideChase,   {  A alt. rows slide out r/l, simult slide in }
                    dmWipe,         {  A individual characters wipe out then in }
                    dmNewWhite,     {  A draws characters one at a time, the new one in white }
                    dmRunLeft,      { MA text runs out then in from the right }
                    dmRunRight,     { MA text runs out then in from the left }
                    dmChaseLeft,    { MA new text chases old text off to the left }
                    dmChaseRight,   { MA new text chases old text off to the right }
                    dmPopUp,        { MA old drops and new rises }
                    dmPaintRandom,  {  A paints dots in random groups }
                    dmPopUpNew,     { MA popup new characters only }
                    dmPopDownNew,   { MA drop-in new characters only }
                    dmPopDown       { MA old rises and new drops }
                );

    EMarquee = Class( Exception );

    TMarquee = Class( TPaintBox )
    private

    isAlive,
    FShowPercentageNumber,
    invalidateBrightOnly,
    FInvalidateDisplay:  Boolean;

    FBrightColour,
    FDimColour:          TColor;

    prevProgress,
    maximum:            Longint;

    FPixelAppearance:   TPixelAppearance;
    myAnimationSpeed:   TAnimationSpeed;

    FCodePage,                      { ANSI code page for character generator }
    FPixelSize,                     { size of a pixel }
    FPixelSizeM1,                   { one less than the size of a pixel - computation speed enhancer }
    FLeftIndent,
    FTopIndent,                     { indentation required to center our text in the paintbox }
    FHorizontalPixels,                      { number of dots across }
    userMaxChars,                   { user's specified # chars across }
    maxCharacters:      Integer;    { number of characters across }

    myAlignment:        TAlignment;

    captionText,
    thisText,
    prevText:           String32;  { previous display characters }

    FWorkRack:          array[ 0..PIXEL_HEIGHT, 0..MAXWIDTH ] of Byte;  { what's going to be displayed }
    FDisplayRack:       array[ 0..PIXEL_HEIGHT, 0..MAXWIDTH ] of Byte;  { what's being displayed }
    FPriorRack:         array[ 0..PIXEL_HEIGHT, 0..MAXWIDTH ] of Byte;  { the last display }

    procedure           AnimationPause;
    procedure           AssembleBits( const s: String );
    procedure           AssembleProgressBits( myProgress: Longint );

    procedure           DrawPixel( col, row: Integer; isDim: Boolean );
    procedure           DrawNormal;
    procedure           DrawSlide;
    procedure           DrawVerticalSlide;
    procedure           DrawSlideChase;
    procedure           DrawWipe;
    procedure           DrawNewWhite;
    procedure           DrawRunLeft;
    procedure           DrawRunRight;
    procedure           DrawChaseLeft;
    procedure           DrawChaseRight;
    procedure           DrawPaintRandom;
    procedure           DrawPopUp;
    procedure           DrawPopDown;
    procedure           DrawPopUpNew;
    procedure           DrawPopDownNew;

    procedure           FillRectangle;
    procedure           Redraw;
    procedure           RecomputeMaxchars;

    protected
    drawSame,                       { if text does not change but a setmessage() happens, should we redraw anyhow? }
    FMechanicalEffect:  Boolean;   { should we use the mechanical effect? (one pos too far then backs up) }
    drawMode:           TDrawMode;

    {   overridden methods }
    procedure           Loaded; override;
    procedure           Paint; override;

    procedure           SetAlignment( how: TAlignment );
    procedure           SetBrightColour( bright: TColor );
    procedure           SetDimColour( dim: TColor );
    procedure           SetPixelSize( pixels: Integer );
    procedure           SetMaxCharacters( messageWidth: Integer );
    procedure           SetMessage( const s: String32 );
    procedure           SetProgress( myProgress: Longint );
    procedure           SetMaxProgress( maxValue: Longint );

    public
    constructor         Create( owner: TComponent ); override;
    destructor          Destroy; override;

    procedure           SetColour( dim, bright: TColor );
    procedure           RepaintMe;
    procedure           Regen;

    published

    property    Alignment:              TAlignment
                                        read myAlignment
                                        write SetAlignment
                                        default taRightJustify;

    property    AnimationSpeed:         TAnimationSpeed
                                        read myAnimationSpeed
                                        write myAnimationSpeed
                                        default anFast;

    property    AnimationStyle:         TDrawMode
                                        read drawMode 
                                        write drawMode
                                        default dmNormal;

    property    BrightColor:            TColor 
                                        read FBrightColour 
                                        write SetBrightColour 
                                        default clAqua;

    property    Caption:                String32 
                                        read captionText 
                                        write SetMessage;

    property    CodePage:               Integer
                                        read FCodepage
                                        write FCodePage
                                        default 0;
                                                                                
    property    DimColor:               TColor 
                                        read FDimColour 
                                        write SetDimColour 
                                        default clNavy;

    property    Enabled:                Boolean
                                        read isAlive
                                        write isAlive
                                        default TRUE;

    property    MaxChars:               Integer 
                                        read userMaxChars 
                                        write SetMaxCharacters 
                                        default MAXCHARS_AUTO;

    property    PipSize:                Integer 
                                        read FPixelSize 
                                        write SetPixelSize 
                                        default 3;

    property    ShowProgressPercent:    Boolean
                                        read FShowPercentageNumber
                                        write FShowPercentageNumber
                                        default TRUE;

    property    Pixel3D:                TPixelAppearance
                                        read FPixelAppearance
                                        write FPixelAppearance
                                        default paFlat;

    property    Progress:               Longint 
                                        read prevProgress
                                        write SetProgress 
                                        default 0;

    property    ProgressMax:            Longint
                                        read maximum 
                                        write SetMaxProgress 
                                        default 100;

    property    RedrawIfUnchanged:      Boolean 
                                        read drawSame 
                                        write drawSame 
                                        default FALSE;

    property    MechanicalEffect:       Boolean 
                                        read FMechanicalEffect 
                                        write FMechanicalEffect 
                                        default FALSE;
    end;


procedure   Register;
procedure   MarqueeDispose;

implementation

uses    WinProcs, Forms, WinTypes;

const

    { $note: this stuff needs to find its way into a .DLL file!!! }

    { The first 128 characters are the same for Western and Easter Europe for now }
    first128Characters: array[ 0..127, 0..PIXEL_HEIGHT ] of Byte =
    (   { $00 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $01 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $02 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $03 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $04 dmnd} ( $00, $00, $00, $04, $0E, $1F, $0E, $04, $00, $00 ),
        { $05 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $06 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $07 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $08 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $09 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $0A '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $0B '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $0C '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $0D '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $0E '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $0F '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $10 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $11 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $12 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $13 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $14 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $15 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $16 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $17 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $18 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $19 '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $1A '-' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $1B 'TL'} ( $00, $00, $1C, $08, $08, $08, $04, $04, $04, $07 ),
        { $1C 'TX'} ( $00, $00, $1C, $08, $08, $08, $05, $02, $05, $00 ),
        { $1D 'CR'} ( $00, $00, $0C, $10, $10, $0E, $05, $06, $05, $05 ),
        { $1E 'PM'} ( $00, $00, $18, $14, $18, $10, $05, $07, $05, $05 ),
        { $1F 'AM'} ( $00, $00, $08, $14, $1C, $14, $05, $07, $05, $05 ),
        { $20 ' ' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $21 '!' } ( $00, $00, $04, $04, $04, $04, $04, $00, $04, $00 ),
        { $22 '"' } ( $00, $00, $0A, $0A, $00, $00, $00, $00, $00, $00 ),
        { $23 '#' } ( $00, $00, $0A, $0A, $1F, $0A, $1F, $0A, $0A, $00 ),
        { $24 '$' } ( $00, $00, $04, $0F, $14, $0E, $05, $1E, $04, $00 ),
        { $25 '%' } ( $00, $00, $18, $19, $02, $04, $08, $13, $03, $00 ),
        { $26 '&' } ( $00, $00, $04, $0F, $14, $0F, $14, $0F, $04, $00 ),
        { $27 ''' } ( $00, $00, $04, $04, $00, $00, $00, $00, $00, $00 ),
        { $28 '(' } ( $00, $00, $02, $04, $08, $08, $08, $08, $04, $02 ),
        { $29 ')' } ( $00, $00, $08, $04, $02, $02, $02, $02, $04, $08 ),
        { $2A '*' } ( $00, $00, $00, $15, $0E, $04, $0E, $15, $00, $00 ),
        { $2B '+' } ( $00, $00, $00, $04, $04, $1F, $04, $04, $00, $00 ),
        { $2C ',' } ( $00, $00, $00, $00, $00, $00, $00, $04, $04, $08 ),
        { $2D '-' } ( $00, $00, $00, $00, $00, $1F, $00, $00, $00, $00 ),
        { $2E '.' } ( $00, $00, $00, $00, $00, $00, $00, $0C, $0C, $00 ),
        { $2F '/' } ( $00, $00, $00, $01, $02, $04, $08, $10, $00, $00 ),
        { $30 '0' } ( $00, $00, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $31 '1' } ( $00, $00, $04, $0C, $14, $04, $04, $04, $1F, $00 ),
        { $32 '2' } ( $00, $00, $0E, $11, $01, $02, $04, $08, $1F, $00 ),
        { $33 '3' } ( $00, $00, $1E, $01, $01, $0E, $01, $01, $1E, $00 ),
        { $34 '4' } ( $00, $00, $12, $12, $12, $1F, $02, $02, $02, $00 ),
        { $35 '5' } ( $00, $00, $1F, $10, $10, $1E, $01, $01, $1E, $00 ),
        { $36 '6' } ( $00, $00, $0E, $10, $10, $1E, $11, $11, $0E, $00 ),
        { $37 '7' } ( $00, $00, $1F, $01, $02, $04, $08, $10, $10, $00 ),
        { $38 '8' } ( $00, $00, $0E, $11, $11, $0E, $11, $11, $0E, $00 ),
        { $39 '9' } ( $00, $00, $0E, $11, $11, $0F, $01, $01, $0E, $00 ),
        { $3A ':' } ( $00, $00, $00, $00, $0C, $0C, $00, $0C, $0C, $00 ),
        { $3B ';' } ( $00, $00, $00, $00, $04, $04, $00, $04, $04, $08 ),
        { $3C '<' } ( $00, $00, $02, $04, $08, $10, $08, $04, $02, $00 ),
        { $3D '=' } ( $00, $00, $00, $00, $1F, $00, $1F, $00, $00, $00 ),
        { $3E '>' } ( $00, $00, $10, $08, $04, $02, $04, $08, $10, $00 ),
        { $3F '?' } ( $00, $00, $0E, $11, $01, $02, $04, $00, $04, $00 ),
        { $40 '@' } ( $00, $00, $0E, $11, $17, $15, $17, $10, $0F, $00 ),
        { $41 'A' } ( $00, $00, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $42 'B' } ( $00, $00, $1E, $11, $11, $1E, $11, $11, $1E, $00 ),
        { $43 'C' } ( $00, $00, $0E, $11, $10, $10, $10, $11, $0E, $00 ),
        { $44 'D' } ( $00, $00, $1E, $09, $09, $09, $09, $09, $1E, $00 ),
        { $45 'E' } ( $00, $00, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $46 'F' } ( $00, $00, $1F, $10, $10, $1E, $10, $10, $10, $00 ),
        { $47 'G' } ( $00, $00, $0E, $10, $10, $17, $11, $11, $0E, $00 ),
        { $48 'H' } ( $00, $00, $11, $11, $11, $1F, $11, $11, $11, $00 ),
        { $49 'I' } ( $00, $00, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $4A 'J' } ( $00, $00, $0F, $01, $01, $01, $01, $11, $0E, $00 ),
        { $4B 'K' } ( $00, $00, $11, $12, $14, $18, $14, $12, $11, $00 ),
        { $4C 'L' } ( $00, $00, $10, $10, $10, $10, $10, $10, $1F, $00 ),
        { $4D 'M' } ( $00, $00, $11, $1B, $15, $11, $11, $11, $11, $00 ),
        { $4E 'N' } ( $00, $00, $11, $19, $15, $13, $11, $11, $11, $00 ),
        { $4F 'O' } ( $00, $00, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $50 'P' } ( $00, $00, $1E, $11, $11, $1E, $10, $10, $10, $00 ),
        { $51 'Q' } ( $00, $00, $0E, $11, $11, $11, $01, $0D, $1E, $01 ),
        { $52 'R' } ( $00, $00, $1E, $11, $11, $1E, $11, $11, $11, $00 ),
        { $53 'S' } ( $00, $00, $0E, $11, $10, $0E, $01, $11, $0E, $00 ),
        { $54 'T' } ( $00, $00, $1F, $04, $04, $04, $04, $04, $04, $00 ),
        { $55 'U' } ( $00, $00, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $56 'V' } ( $00, $00, $11, $11, $11, $11, $0A, $0A, $04, $00 ),
        { $57 'W' } ( $00, $00, $11, $11, $11, $11, $15, $15, $0A, $00 ),
        { $58 'X' } ( $00, $00, $11, $11, $0A, $04, $0A, $11, $11, $00 ),
        { $59 'Y' } ( $00, $00, $11, $11, $0A, $04, $04, $04, $04, $00 ),
        { $5A 'Z' } ( $00, $00, $1F, $01, $02, $04, $08, $10, $1F, $00 ),
        { $5B '[' } ( $00, $00, $0E, $08, $08, $08, $08, $08, $0E, $00 ),
        { $5C '\' } ( $00, $00, $00, $10, $08, $04, $02, $01, $00, $00 ),
        { $5D ']' } ( $00, $00, $0E, $02, $02, $02, $02, $02, $0E, $00 ),
        { $5E '^' } ( $00, $00, $04, $0A, $11, $00, $00, $00, $00, $00 ),
        { $5F '_' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $1F ),
        { $60 '`' } ( $00, $00, $04, $02, $01, $00, $00, $00, $00, $00 ),
        { $61 'a' } ( $00, $00, $00, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $62 'b' } ( $00, $00, $10, $10, $1E, $11, $11, $11, $1E, $00 ),
        { $63 'c' } ( $00, $00, $00, $00, $0F, $10, $10, $10, $0F, $00 ),
        { $64 'd' } ( $00, $00, $01, $01, $0F, $11, $11, $11, $0F, $00 ),
        { $65 'e' } ( $00, $00, $00, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $66 'f' } ( $00, $00, $07, $08, $08, $1E, $08, $08, $08, $00 ),
        { $67 'g' } ( $00, $00, $00, $00, $0F, $11, $11, $0F, $01, $0E ),
        { $68 'h' } ( $00, $00, $10, $10, $10, $1E, $11, $11, $11, $00 ),
        { $69 'i' } ( $00, $00, $00, $04, $00, $0C, $04, $04, $0E, $00 ),
        { $6A 'j' } ( $00, $00, $00, $04, $00, $0C, $04, $04, $04, $0C ),
        { $6B 'k' } ( $00, $00, $10, $10, $12, $14, $18, $14, $12, $00 ),
        { $6C 'l' } ( $00, $00, $0C, $04, $04, $04, $04, $04, $0E, $00 ),
        { $6D 'm' } ( $00, $00, $00, $00, $1A, $15, $15, $15, $15, $00 ),
        { $6E 'n' } ( $00, $00, $00, $00, $16, $19, $11, $11, $11, $00 ),
        { $6F 'o' } ( $00, $00, $00, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $70 'p' } ( $00, $00, $00, $00, $1E, $11, $11, $1E, $10, $10 ),
        { $71 'q' } ( $00, $00, $00, $00, $0F, $11, $11, $0F, $01, $01 ),
        { $72 'r' } ( $00, $00, $00, $00, $0B, $0C, $08, $08, $08, $00 ),
        { $73 's' } ( $00, $00, $00, $00, $0F, $10, $0E, $01, $1E, $00 ),
        { $74 't' } ( $00, $00, $00, $04, $0E, $04, $04, $04, $06, $00 ),
        { $75 'u' } ( $00, $00, $00, $00, $11, $11, $11, $11, $0F, $00 ),
        { $76 'v' } ( $00, $00, $00, $00, $11, $11, $0A, $0A, $04, $00 ),
        { $77 'w' } ( $00, $00, $00, $00, $11, $15, $15, $15, $0A, $00 ),
        { $78 'x' } ( $00, $00, $00, $00, $11, $0A, $04, $0A, $11, $00 ),
        { $79 'y' } ( $00, $00, $00, $00, $11, $11, $11, $0F, $01, $0E ),
        { $7A 'z' } ( $00, $00, $00, $00, $1F, $02, $04, $08, $1F, $00 ),
        { $7B oBr } ( $00, $00, $02, $04, $04, $08, $04, $04, $02, $00 ),
        { $7C '|' } ( $04, $04, $04, $04, $04, $04, $04, $04, $04, $04 ),
        { $7D cBr } ( $00, $00, $08, $04, $04, $02, $04, $04, $08, $00 ),
        { $7E '~' } ( $00, $00, $1F, $10, $00, $00, $00, $00, $00, $00 ),
        { $7F ' ' } ( $00, $00, $1F, $1F, $1F, $1F, $1F, $1F, $1F, $1F )
    );

    
    { Code Page 1250 is for Eastern Europe }
    CodePage1250:   array[ 0..127, 0..PIXEL_HEIGHT ] of Byte =
    (
        { $80 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $81 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $82 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $02, $04 ),
        { $83 '?' } ( $00, $03, $04, $04, $0E, $04, $04, $04, $18, $00 ),
        { $84 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $05, $0A ),
        { $85 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $15, $00 ),
        { $86 '?' } ( $00, $00, $04, $04, $1F, $04, $04, $04, $04, $00 ),
        { $87 '?' } ( $00, $00, $04, $04, $1F, $04, $1F, $04, $04, $00 ),
        { $88 '?' } ( $04, $0A, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $89 '?' } ( $00, $00, $19, $1A, $04, $08, $10, $1B, $1B, $00 ),
        { $8A '?' } ( $0A, $04, $0E, $11, $10, $0E, $01, $11, $0E, $00 ),
        { $8B '?' } ( $00, $00, $00, $02, $04, $08, $04, $02, $00, $00 ),
        { $8C '?' } ( $02, $04, $0E, $11, $10, $0E, $01, $11, $0E, $00 ),
        { $8D '?' } ( $0A, $04, $1F, $04, $04, $04, $04, $04, $04, $00 ),
        { $8E '?' } ( $0A, $04, $1F, $01, $02, $04, $08, $10, $1F, $00 ),
        { $8F '?' } ( $02, $04, $1F, $01, $02, $04, $08, $10, $1F, $00 ),
        { $90 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $91 '?' } ( $00, $08, $04, $00, $00, $00, $00, $00, $00, $00 ),
        { $92 '?' } ( $00, $04, $08, $00, $00, $00, $00, $00, $00, $00 ),
        { $93 '?' } ( $00, $0A, $05, $00, $00, $00, $00, $00, $00, $00 ),
        { $94 '?' } ( $00, $05, $0A, $00, $00, $00, $00, $00, $00, $00 ),
        { $95 '?' } ( $00, $00, $0C, $0C, $00, $00, $00, $00, $00, $00 ),
        { $96 '?' } ( $00, $00, $00, $00, $00, $0E, $00, $00, $00, $00 ),
        { $97 '?' } ( $00, $00, $00, $00, $00, $1F, $00, $00, $00, $00 ),
        { $98 '?' } ( $00, $09, $16, $00, $00, $00, $00, $00, $00, $00 ),
        { $99 '?' } ( $00, $1C, $08, $07, $05, $00, $00, $00, $00, $00 ),
        { $9A '?' } ( $00, $0A, $04, $00, $0F, $10, $0E, $01, $1E, $00 ),
        { $9B '?' } ( $00, $00, $00, $08, $04, $02, $04, $08, $00, $00 ),
        { $9C '?' } ( $00, $02, $04, $00, $0F, $10, $0E, $01, $1E, $00 ),
        { $9D '?' } ( $00, $00, $01, $05, $0E, $04, $04, $04, $06, $00 ),
        { $9E '?' } ( $00, $0A, $04, $00, $1F, $02, $04, $08, $1F, $00 ),
        { $9F '?' } ( $00, $02, $04, $00, $1F, $02, $04, $08, $1F, $00 ),
        { $A0 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $A1 '?' } ( $0A, $04, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $A2 '?' } ( $09, $06, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $A3 '?' } ( $00, $00, $08, $08, $0C, $08, $18, $08, $0F, $00 ),
        { $A4 '?' } ( $00, $00, $00, $11, $0E, $0A, $0E, $11, $00, $00 ),
        { $A5 '?' } ( $00, $00, $0E, $11, $11, $1F, $11, $11, $11, $03 ),
        { $A6 '?' } ( $04, $04, $04, $04, $00, $00, $04, $04, $04, $04 ),
        { $A7 '?' } ( $00, $00, $0E, $09, $14, $0A, $05, $12, $0E, $00 ),
        { $A8 '?' } ( $0A, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $A9 '?' } ( $00, $0E, $11, $17, $19, $17, $11, $0E, $00, $00 ),
        { $AA '?' } ( $00, $00, $0E, $11, $10, $0E, $01, $11, $0E, $0C ),
        { $AB '?' } ( $00, $00, $05, $0A, $14, $0A, $05, $00, $00, $00 ),
        { $AC '?' } ( $00, $00, $00, $00, $00, $1F, $01, $00, $00, $00 ),
        { $AD '?' } ( $00, $00, $00, $00, $00, $0F, $00, $00, $00, $00 ),
        { $AE '?' } ( $00, $0E, $11, $17, $19, $19, $11, $0E, $00, $00 ),
        { $AF '?' } ( $04, $00, $1F, $01, $02, $04, $08, $10, $1F, $00 ),
        { $B0 '?' } ( $04, $0A, $04, $00, $00, $00, $00, $00, $00, $00 ),
        { $B1 '?' } ( $00, $00, $04, $0E, $04, $00, $0E, $00, $00, $00 ),
        { $B2 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $04, $06 ),
        { $B3 '?' } ( $00, $00, $0C, $04, $06, $04, $0C, $04, $0E, $00 ),
        { $B4 '?' } ( $02, $04, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $B5 '?' } ( $00, $00, $00, $00, $12, $12, $12, $12, $1F, $10 ),
        { $B6 '?' } ( $00, $00, $1F, $1A, $1A, $0A, $0A, $0A, $1B, $00 ),
        { $B7 '?' } ( $00, $00, $00, $00, $00, $04, $00, $00, $00, $00 ),
        { $B8 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $02, $06 ),
        { $B9 '?' } ( $00, $00, $00, $00, $0E, $01, $0F, $11, $0F, $03 ),
        { $BA '?' } ( $00, $00, $00, $00, $0F, $10, $0E, $01, $1E, $0C ),
        { $BB '?' } ( $00, $00, $14, $0A, $05, $0A, $14, $00, $00, $00 ),
        { $BC '?' } ( $00, $00, $12, $14, $10, $10, $10, $10, $1F, $00 ),
        { $BD '?' } ( $05, $0A, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $BE '?' } ( $00, $00, $0D, $05, $04, $04, $04, $04, $0E, $00 ),
        { $BF '?' } ( $00, $00, $04, $00, $1F, $02, $04, $08, $1F, $00 ),
        { $C0 '?' } ( $02, $04, $1E, $11, $11, $1E, $11, $11, $11, $00 ),
        { $C1 '?' } ( $02, $04, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C2 '?' } ( $04, $0A, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C3 '?' } ( $0A, $04, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C4 '?' } ( $0A, $00, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C5 '?' } ( $02, $04, $10, $10, $10, $10, $10, $10, $1F, $00 ),
        { $C6 '?' } ( $02, $04, $0E, $11, $10, $10, $10, $11, $0E, $00 ),
        { $C7 '?' } ( $00, $00, $0E, $11, $10, $10, $10, $11, $0E, $0C ),
        { $C8 '?' } ( $0A, $04, $0E, $11, $10, $10, $10, $11, $0E, $00 ),
        { $C9 '?' } ( $02, $04, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $CA '?' } ( $00, $00, $1F, $10, $10, $1E, $10, $10, $1F, $03 ),
        { $CB '?' } ( $0A, $00, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $CC '?' } ( $0A, $04, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $CD '?' } ( $02, $04, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $CE '?' } ( $04, $0A, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $CF '?' } ( $0A, $04, $1E, $09, $09, $09, $09, $09, $1E, $00 ),
        { $D0 '?' } ( $00, $00, $1E, $09, $09, $1D, $09, $09, $1E, $00 ),
        { $D1 '?' } ( $02, $04, $11, $19, $15, $13, $11, $11, $11, $00 ),
        { $D2 '?' } ( $0A, $04, $11, $19, $15, $13, $11, $11, $11, $00 ),
        { $D3 '?' } ( $02, $04, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D4 '?' } ( $04, $0A, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D5 '?' } ( $05, $0A, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D6 '?' } ( $0A, $00, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D7 '?' } ( $00, $00, $00, $11, $0A, $04, $0A, $11, $00, $00 ),
        { $D8 '?' } ( $0A, $04, $1E, $11, $11, $1E, $11, $11, $11, $00 ),
        { $D9 '?' } ( $04, $0A, $15, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DA '?' } ( $02, $04, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DB '?' } ( $05, $0A, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DC '?' } ( $0A, $00, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DD '?' } ( $02, $04, $11, $11, $0A, $04, $04, $04, $04, $00 ),
        { $DE '?' } ( $00, $00, $1F, $04, $04, $04, $04, $04, $04, $0C ),
        { $DF '?' } ( $00, $00, $0C, $12, $12, $1E, $11, $11, $1E, $10 ),
        { $E0 '?' } ( $00, $02, $04, $00, $0B, $0C, $08, $08, $08, $00 ),
        { $E1 '?' } ( $00, $02, $04, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E2 '?' } ( $00, $04, $0A, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E3 '?' } ( $00, $0A, $04, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E4 '?' } ( $00, $00, $0A, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E5 '?' } ( $02, $04, $0C, $04, $04, $04, $04, $04, $0E, $00 ),
        { $E6 '?' } ( $00, $02, $04, $00, $0F, $10, $10, $10, $0F, $00 ),
        { $E7 '?' } ( $00, $00, $00, $00, $0F, $10, $10, $10, $0F, $0C ),
        { $E8 '?' } ( $00, $0A, $04, $00, $0F, $10, $10, $10, $0F, $00 ),
        { $E9 '?' } ( $00, $02, $04, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $EA '?' } ( $00, $00, $00, $00, $0E, $11, $1F, $10, $0F, $06 ),
        { $EB '?' } ( $00, $00, $0A, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $EC '?' } ( $00, $0A, $04, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $ED '?' } ( $00, $02, $04, $00, $0C, $04, $04, $04, $0E, $00 ),
        { $EE '?' } ( $00, $04, $0A, $00, $0C, $04, $04, $04, $0E, $00 ),
        { $EF '?' } ( $0A, $04, $01, $01, $0F, $11, $11, $11, $0F, $00 ),
        { $F0 '?' } ( $00, $00, $02, $07, $02, $0E, $12, $12, $0E, $00 ),
        { $F1 '?' } ( $00, $02, $04, $00, $16, $19, $11, $11, $11, $00 ),
        { $F2 '?' } ( $00, $0A, $04, $00, $16, $19, $11, $11, $11, $00 ),
        { $F3 '?' } ( $00, $02, $04, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F4 '?' } ( $00, $04, $0A, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F5 '?' } ( $00, $05, $0A, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F6 '?' } ( $00, $00, $0A, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F7 '?' } ( $00, $00, $00, $04, $00, $1F, $00, $04, $00, $00 ),
        { $F8 '?' } ( $00, $0A, $04, $00, $0B, $0C, $08, $08, $08, $00 ),
        { $F9 '?' } ( $00, $04, $0A, $04, $11, $11, $11, $11, $0F, $00 ),
        { $FA '?' } ( $00, $02, $04, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FB '?' } ( $00, $05, $0A, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FC '?' } ( $00, $00, $0A, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FD '?' } ( $00, $02, $04, $00, $11, $11, $11, $0F, $01, $0E ),
        { $FE '?' } ( $00, $00, $00, $04, $0E, $04, $04, $04, $06, $0C ),
        { $FF '?' } ( $00, $04, $00, $00, $00, $00, $00, $00, $00, $00 )
    );

    { Code Page 1252 is for Western Europe }
    codePage1252: array[ 0..127, 0..PIXEL_HEIGHT ] of Byte =
    (   { $80 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $81 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $82 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $02, $04 ),
        { $83 '?' } ( $00, $03, $04, $04, $0E, $04, $04, $04, $18, $00 ),
        { $84 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $05, $0A ),
        { $85 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $15, $00 ),
        { $86 '?' } ( $00, $00, $04, $04, $1F, $04, $04, $04, $04, $00 ),
        { $87 '?' } ( $00, $00, $04, $04, $1F, $04, $1F, $04, $04, $00 ),
        { $88 '?' } ( $04, $0A, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $89 '?' } ( $00, $00, $19, $1A, $04, $08, $10, $1B, $1B, $00 ),
        { $8A '?' } ( $0A, $04, $0E, $11, $10, $0E, $01, $11, $0E, $00 ),
        { $8B '?' } ( $00, $00, $00, $02, $04, $08, $04, $02, $00, $00 ),
        { $8C '?' } ( $00, $00, $0F, $14, $14, $16, $14, $14, $0F, $00 ),
        { $8D '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $8E '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $8F '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $90 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $91 '?' } ( $00, $08, $04, $00, $00, $00, $00, $00, $00, $00 ),
        { $92 '?' } ( $00, $04, $08, $00, $00, $00, $00, $00, $00, $00 ),
        { $93 '?' } ( $00, $0A, $05, $00, $00, $00, $00, $00, $00, $00 ),
        { $94 '?' } ( $00, $05, $0A, $00, $00, $00, $00, $00, $00, $00 ),
        { $95 '?' } ( $00, $00, $00, $0C, $0C, $00, $00, $00, $00, $00 ),
        { $96 '?' } ( $00, $00, $00, $00, $00, $0E, $00, $00, $00, $00 ),
        { $97 '?' } ( $00, $00, $00, $00, $00, $1F, $00, $00, $00, $00 ),
        { $98 '?' } ( $00, $09, $16, $00, $00, $00, $00, $00, $00, $00 ),
        { $99 '?' } ( $00, $1C, $08, $07, $05, $00, $00, $00, $00, $00 ),
        { $9A '?' } ( $00, $0A, $04, $00, $0F, $10, $0E, $01, $1E, $00 ),
        { $9B '?' } ( $00, $00, $00, $08, $04, $02, $04, $08, $00, $00 ),
        { $9C '?' } ( $00, $00, $00, $00, $0A, $15, $16, $14, $0B, $00 ),
        { $9D '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $9E '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $9F '?' } ( $0A, $00, $11, $11, $0A, $04, $04, $04, $04, $00 ),
        { $A0 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $A1 '?' } ( $00, $00, $04, $00, $04, $04, $04, $04, $04, $00 ),
        { $A2 '?' } ( $00, $00, $00, $04, $0F, $14, $14, $14, $0F, $04 ),
        { $A3 '?' } ( $00, $00, $04, $0A, $08, $1C, $08, $08, $0F, $00 ),
        { $A4 '?' } ( $00, $00, $00, $11, $0E, $0A, $0E, $11, $00, $00 ),
        { $A5 '?' } ( $00, $00, $11, $11, $0A, $1F, $04, $1F, $04, $00 ),
        { $A6 '?' } ( $04, $04, $04, $04, $00, $00, $04, $04, $04, $04 ),
        { $A7 '?' } ( $00, $00, $0E, $09, $14, $0A, $05, $12, $0E, $00 ),
        { $A8 '?' } ( $0A, $00, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $A9 '?' } ( $00, $0E, $11, $17, $19, $17, $11, $0E, $00, $00 ),
        { $AA '?' } ( $00, $0C, $02, $0E, $12, $0E, $00, $00, $00, $00 ),
        { $AB '?' } ( $00, $00, $05, $0A, $14, $0A, $05, $00, $00, $00 ),
        { $AC '?' } ( $00, $00, $00, $00, $00, $1F, $01, $00, $00, $00 ),
        { $AD '?' } ( $00, $00, $00, $00, $00, $0F, $00, $00, $00, $00 ),
        { $AE '?' } ( $00, $0E, $11, $17, $19, $19, $11, $0E, $00, $00 ),
        { $AF '?' } ( $00, $0E, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $B0 '?' } ( $04, $0A, $04, $00, $00, $00, $00, $00, $00, $00 ),
        { $B1 '?' } ( $00, $00, $04, $0E, $04, $00, $0E, $00, $00, $00 ),
        { $B2 '?' } ( $00, $0E, $02, $0E, $08, $0E, $00, $00, $00, $00 ),
        { $B3 '?' } ( $00, $0E, $02, $06, $02, $0E, $00, $00, $00, $00 ),
        { $B4 '?' } ( $02, $04, $00, $00, $00, $00, $00, $00, $00, $00 ),
        { $B5 '?' } ( $00, $00, $00, $00, $12, $12, $12, $12, $1F, $10 ),
        { $B6 '?' } ( $00, $00, $1F, $1A, $1A, $0A, $0A, $0A, $1B, $00 ),
        { $B7 '?' } ( $00, $00, $00, $00, $00, $04, $00, $00, $00, $00 ),
        { $B8 '?' } ( $00, $00, $00, $00, $00, $00, $00, $00, $02, $06 ),
        { $B9 '?' } ( $00, $0C, $04, $04, $04, $04, $00, $00, $00, $00 ),
        { $BA '?' } ( $00, $04, $0A, $0A, $0A, $04, $00, $00, $00, $00 ),
        { $BB '?' } ( $00, $00, $14, $0A, $05, $0A, $14, $00, $00, $00 ),
        { $BC '?' } ( $00, $18, $08, $09, $0A, $0E, $0A, $13, $01, $01 ),
        { $BD '?' } ( $00, $18, $08, $09, $0A, $0F, $09, $13, $02, $03 ),
        { $BE '?' } ( $00, $18, $08, $19, $0A, $1E, $0A, $13, $01, $01 ),
        { $BF '?' } ( $00, $00, $04, $00, $04, $08, $10, $11, $0E, $00 ),
        { $C0 '?' } ( $08, $04, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C1 '?' } ( $02, $04, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C2 '?' } ( $04, $0A, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C3 '?' } ( $05, $0A, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C4 '?' } ( $0A, $00, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C5 '?' } ( $04, $00, $0E, $11, $11, $1F, $11, $11, $11, $00 ),
        { $C6 '?' } ( $00, $00, $0F, $14, $14, $1E, $14, $14, $17, $00 ),
        { $C7 '?' } ( $00, $00, $0E, $11, $10, $10, $10, $11, $0E, $0C ),
        { $C8 '?' } ( $08, $04, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $C9 '?' } ( $02, $04, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $CA '?' } ( $04, $0A, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $CB '?' } ( $0A, $00, $1F, $10, $10, $1E, $10, $10, $1F, $00 ),
        { $CC '?' } ( $08, $04, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $CD '?' } ( $02, $04, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $CE '?' } ( $04, $0A, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $CF '?' } ( $0A, $00, $0E, $04, $04, $04, $04, $04, $0E, $00 ),
        { $D0 '?' } ( $00, $00, $1E, $09, $09, $1D, $09, $09, $1E, $00 ),
        { $D1 '?' } ( $05, $0A, $11, $19, $15, $13, $11, $11, $11, $00 ),
        { $D2 '?' } ( $08, $04, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D3 '?' } ( $02, $04, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D4 '?' } ( $04, $0A, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D5 '?' } ( $05, $0A, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D6 '?' } ( $0A, $00, $0E, $11, $11, $11, $11, $11, $0E, $00 ),
        { $D7 '?' } ( $00, $00, $00, $11, $0A, $04, $0A, $11, $00, $00 ),
        { $D8 '?' } ( $00, $01, $0E, $13, $15, $15, $15, $19, $0E, $10 ),
        { $D9 '?' } ( $08, $04, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DA '?' } ( $02, $04, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DB '?' } ( $04, $0A, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DC '?' } ( $0A, $00, $11, $11, $11, $11, $11, $11, $0E, $00 ),
        { $DD '?' } ( $02, $04, $11, $11, $0A, $04, $04, $04, $04, $00 ),
        { $DE '?' } ( $00, $00, $10, $1C, $12, $12, $12, $1C, $10, $00 ),
        { $DF '?' } ( $00, $00, $0C, $12, $12, $1E, $11, $11, $1E, $10 ),
        { $E0 '?' } ( $00, $08, $04, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E1 '?' } ( $00, $02, $04, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E2 '?' } ( $00, $04, $0A, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E3 '?' } ( $00, $05, $0A, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E4 '?' } ( $00, $00, $0A, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E5 '?' } ( $04, $0A, $04, $00, $0E, $01, $0F, $11, $0F, $00 ),
        { $E6 '?' } ( $00, $00, $00, $00, $0F, $14, $1E, $14, $17, $00 ),
        { $E7 '?' } ( $00, $00, $00, $00, $0F, $10, $10, $10, $0F, $0C ),
        { $E8 '?' } ( $00, $08, $04, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $E9 '?' } ( $00, $02, $04, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $EA '?' } ( $00, $04, $0A, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $EB '?' } ( $00, $00, $0A, $00, $0E, $11, $1F, $10, $0F, $00 ),
        { $EC '?' } ( $00, $08, $04, $00, $0C, $04, $04, $04, $0E, $00 ),
        { $ED '?' } ( $00, $02, $04, $00, $0C, $04, $04, $04, $0E, $00 ),
        { $EE '?' } ( $00, $04, $0A, $00, $0C, $04, $04, $04, $0E, $00 ),
        { $EF '?' } ( $00, $00, $0A, $00, $0C, $04, $04, $04, $0E, $00 ),
        { $F0 '?' } ( $00, $00, $02, $07, $02, $0E, $12, $12, $0C, $00 ),
        { $F1 '?' } ( $00, $05, $0A, $00, $16, $19, $11, $11, $11, $00 ),
        { $F2 '?' } ( $00, $08, $04, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F3 '?' } ( $00, $02, $04, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F4 '?' } ( $00, $04, $0A, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F5 '?' } ( $00, $05, $0A, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F6 '?' } ( $00, $00, $0A, $00, $0E, $11, $11, $11, $0E, $00 ),
        { $F7 '?' } ( $00, $00, $00, $04, $00, $1F, $00, $04, $00, $00 ),
        { $F8 '?' } ( $00, $00, $00, $01, $0E, $13, $15, $19, $0E, $10 ),
        { $F9 '?' } ( $00, $08, $04, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FA '?' } ( $00, $02, $04, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FB '?' } ( $00, $04, $0A, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FC '?' } ( $00, $00, $0A, $00, $11, $11, $11, $11, $0F, $00 ),
        { $FD '?' } ( $00, $02, $04, $00, $11, $11, $11, $0F, $01, $0E ),
        { $FE '?' } ( $00, $00, $10, $1E, $11, $11, $11, $11, $1E, $10 ),
        { $FF '?' } ( $00, $00, $0A, $00, $11, $11, $11, $0F, $01, $0E )
    );

type
    TIntArray   = array[ 0 .. ( MAXWIDTH * ( PIXEL_HEIGHT + 1 ) ) ] of Integer;
    
    TBitPattern = Record
        bit:   array[ 0..255, 0..PIXEL_HEIGHT ] of Byte;
    end;

    pBitPattern = ^TBitPattern;

var
    bitPatternBuilt,                { have we built the bit patterns yet? }
    delayInstance:      Boolean;
    bitPattern:         pBitPattern;

procedure Register;

{   VCL Registration function.
}

begin
    RegisterComponents( 'Midnight', [ TMarquee ] )
end;


destructor TMarquee.Destroy;
begin
    inherited Destroy
end;


procedure TMarquee.RecomputeMaxchars;

{   If the user has specified an automatic # chars across, we need to
    compute the actual number of characters when the form is created,
    resized or the size of the pixels is changed.

    inputs:
    none

    outputs:
    none

    returns:
    none
}

var
    oldMaxchars,
    oldLeftIndent,
    oldTopIndent:   Integer;

begin
    if ( userMaxChars = MAXCHARS_AUTO ) then
    begin
        if ( FPixelSize < 2 ) then
        begin
            FPixelSize := 2;
            FPixelSizeM1 := 1
        end;

        maxCharacters := Width div ( 6 * FPixelSize );
        if ( maxCharacters > MAX_CHARS ) then
            maxCharacters := MAX_CHARS;

        if ( maxCharacters < 0 ) then
            maxCharacters := 0
    end;

    { while we're here, recompute the FLeftIndentation value }

    FLeftIndent := 1 + ( ( Width - ( 6 * FPixelSize * maxCharacters ) ) div 2 );
    if ( FLeftIndent < 0 ) then
        FLeftIndent := 0;

    FHorizontalPixels := 6 * maxCharacters;

    { and let us not forget the top indentation! }
    FTopIndent := 1 + ( ( Height - ( ( PIXEL_HEIGHT + 1 ) * FPixelSize ) ) div 2 );
    if ( FTopIndent < 0 ) then
        FTopIndent := 0
end;


procedure TMarquee.SetMaxProgress( maxValue: Longint );

{   If the marquis is being used as a progress meter, this function is
    used to set the maximum progress value

    inputs:
    maxValue    value equal to 100% of progress

    outputs:
    none

    returns:
    none
}

begin
    prevProgress := -1;
    maximum := maxValue
end;


procedure TMarquee.SetMaxCharacters( messageWidth: Integer );

{   define the maximum width of the marquis, in text characters.

    inputs:
    messageWidth    number of characters, or 0 to base on frame's width

    outputs:
    none

    returns:
    none
}

var
    i,
    j:  Integer;

begin
    maxCharacters := messageWidth;
    userMaxChars := messageWidth;
    if ( FPixelSize < 2 ) then
    begin
        FPixelSize := 2;
        FPixelSizeM1 := 1
    end;

    RecomputeMaxChars;

    for i := 0 to PIXEL_HEIGHT do
        for j := 0 to MAXWIDTH do
        begin
            FWorkRack[ i, j ] := 0;
            FDisplayRack[ i, j ] := 0;
            FPriorRack[ i, j ] := 0
        end;

    FInvalidateDisplay := TRUE;
    DrawNormal
end;


procedure MarqueeDispose;
begin
    if ( bitPatternBuilt ) then
    begin
        dispose( bitPattern );
        bitPatternBuilt := FALSE
    end;
end;


procedure TMarquee.Loaded;

{   After we know what code page the user wants, if it's default, set it here.
}

var
    i,
    j:  Integer;
    
begin
    if ( not bitPatternBuilt ) then
    begin
        if ( FCodePage = 0 ) then
{$IFDEF WIN32}
            FCodePage := GetACP;       { retrieve the ANSI Code Page }
{$ELSE}
            FCodePage := 1252;         { Delphi1 - Sorry, only U.S. code page can be automagically set :-( }
{$ENDIF}

        { first 128 characters are the same for all codepages -- so far }
        for i := 0 to 127 do
            for j := 0 to PIXEL_HEIGHT do
                bitPattern^.bit[ i, j ] := first128Characters[ i, j ];

        if ( FCodePage = 1250 ) then   { eastern european codepage 1250 }
        begin
            for i := 0 to 127 do
                for j := 0 to PIXEL_HEIGHT do
                    bitPattern^.bit[ i + 128, j ] := codePage1250[ i, j ]
        end
        else        { western european codepage 1252 }
        begin
            for i := 0 to 127 do
                for j := 0 to PIXEL_HEIGHT do
                    bitPattern^.bit[ i + 128, j ] := codePage1252[ i, j ]
        end;

        bitPatternBuilt := TRUE
    end
end;


constructor TMarquee.Create( owner: TComponent );

{   Default VCL constructor
}

var
    i,
    j:          Integer;

begin
    inherited Create( owner );
 
    { establish defaults:
        FMechanicalEffect:   NO
        drawSame:           NO
        FPixelSize:          3
        drawMode:           Normal
        colour:             Aqua dots on Navy background
        enabled:            TRUE
    }

    { allocate memory for the pattern buffer the first time }
    if ( not bitPatternBuilt ) then
        new( bitPattern );
        
    myAnimationSpeed := anFast;
    myAlignment := taRightJustify;
    isAlive := TRUE;
    FPixelAppearance := paFlat;
    Color := clBlack;
    maximum := 0;
    FPixelSize := 3;
    FPixelSizeM1 := 2;
    FCodePage := 0;
    Canvas.Pen.Width := 1;
    canvas.Brush.Style := bsSolid;
    
    FMechanicalEffect := FALSE;
    FShowPercentageNumber := TRUE;

    drawSame := FALSE;
    userMaxChars := MAXCHARS_AUTO;
    RecomputeMaxchars;

    prevText := '';
    drawMode := dmNormal;
    FBrightColour := clAqua;
    FDimColour := clNavy;

    Canvas.Pen.Mode := pmCopy;

    invalidateBrightOnly := FALSE;
    FInvalidateDisplay := TRUE;
    for i := 0 to PIXEL_HEIGHT do
        for j := 0 to MAXWIDTH do
        begin
            FWorkRack[ i, j ] := 0;
            FDisplayRack[ i, j ] := 0;
            FPriorRack[ i, j ] := 0
        end
end;


procedure TMarquee.DrawPaintRandom;

{   Draws the marquee by randomly filling in the zots.
}

    procedure Shuffle( var aList: TIntArray; aLength: Integer );

    {   Procedure internal to DrawPaintRandom that shuffles the order of the
        bits.
    }
    
    var
        stopAt,                     { stop random assignment at this point }
        source,
        dest,                     
        i:          Integer;
        tempList:   TIntArray;
        
    begin
        { move the items from aList to tempList; reset the items in the original list }

        for i := 0 to aLength - 1 do
        begin
            tempList[ i ] := aList[ i ];
            aList[ i ] := -1
        end;

        { move the items from tempList back to aList randomly }

        stopAt := aLength div 50;
        
        for i := 0 to ( aLength - 1 ) - stopAt do
        begin
            repeat
                dest := Random( aLength );
            until ( aList[ dest ] = -1 );
            
            aList[ dest ] := tempList[ i ]
        end;

        {   We fudge things a little by stopping the shuffle when 1/50th of the
            bits remain and assign the balance of the bits (just a few) linearly.
            We do this because at this point, the probability of the random number
            generator selecting a novel destination is 50 to 1, which means it would
            otherwise spend a lot of time trying to randomly assign the last few
            cells.
        }

        if ( stopAt < 1 ) then
            Exit;
            
        for source := ( aLength - 1 ) - stopAt to aLength - 1 do
            for dest := 0 to aLength - 1 do
                if ( aList[ dest ] = -1 ) then
                begin
                    aList[ dest ] := tempList[ source ];
                    break
                end
    end;

var
    myRow,
    myCol,
    i:      Integer;
    list:   TIntArray;

begin
    for i := 0 to FHorizontalPixels * ( PIXEL_HEIGHT + 1 ) do
        list[ i ] := i;
    
    Shuffle( list, FHorizontalPixels * ( PIXEL_HEIGHT + 1 ) + 1 );

    for i := 0 to FHorizontalPixels * ( PIXEL_HEIGHT + 1 ) - 1 do
    begin
        myRow := ( list[ i ] - 1 ) div FHorizontalPixels;
        myCol := ( list[ i ] - 1 ) mod FHorizontalPixels;

        FDisplayRack[ myRow, myCol ] := FWorkRack[ myRow, myCol ];

        if ( ( i mod FHorizontalPixels ) = ( FHorizontalPixels - 1 ) ) then
        begin
            AnimationPause;
            Redraw
        end
    end
end;


procedure TMarquee.AssembleProgressBits( myProgress: Longint );

{   Assembles bits for the progress message.
}

var
    percentage,
    progressBits,
    firstPixel,
    character,
    myPlace,
    len,
    row,
    col:            Integer;

    myText:         String[ 4 ];

    myByte:         Byte;

begin
    if ( maximum < 1 ) then
        Exit;

    { calculate how far the highlight goes across }
    progressBits := ( FHorizontalPixels * myProgress ) div maximum;
    percentage := ( 100 * myProgress ) div maximum;

    if ( percentage < 0 ) then
        percentage := 0
    else if ( percentage > 100 ) then
        percentage := 100;

    { paint the highlight - uniform line }
    for col := 0 to FHorizontalPixels - 1 do
        if ( col <= progressBits ) then
        begin
            for row := 0 to PIXEL_HEIGHT do
                FWorkRack[ row, col ] := 1     { either 1 or 0 }
        end
        else
        begin
            for row := 0 to PIXEL_HEIGHT do
                FWorkRack[ row, col ] := 0
        end;

    { user may only want a bar with no number in the middle }
    if ( FShowPercentageNumber ) then
    begin
        myText := Format( '%2.2d%%', [ percentage ] );

        { center the text [n[n]]n% in the box and flip its bits }

        { calculate first pixel to draw in to center text }
        len := Length( myText );
        firstPixel := ( FHorizontalPixels - ( len * 6 ) ) div 2;

        for character := 1 to len do
            for row := 0 to PIXEL_HEIGHT do
            begin
                myByte := bitPattern^.bit[ ord( myText[ character ] ), row ];
                myPlace := character * 6 + firstPixel;

                for col := 0 to 4 do
                begin
                    if ( ( myByte AND $01 ) <> 0 ) then
                        FWorkRack[ row, myPlace ] := FWorkRack[ row, myPlace ] XOR $01;

                    myByte := myByte SHR 1;
                    Dec( myPlace )
                end
            end
    end
end;

procedure TMarquee.AssembleBits( const s: String );

{ copies bitPatterns to FWorkRack.
}

var
    myByte:     Byte;

    myPlace,
    row,
    len,
    col,
    character:  Integer;

begin
    { clear the work rack }
    for row := 0 to PIXEL_HEIGHT do
        for col := 0 to MAXWIDTH do
            FWorkRack[ row, col ] := 0;

    len := Length( s );
    for character := 1 to len do
        for row := 0 to PIXEL_HEIGHT do
        begin
            myByte := bitPattern^.bit[ ord( s[ character ] ), row ];
            myPlace := character * 6 - 2;

            for col := 0 to 4 do
            begin
                FWorkRack[ row, myPlace ] := myByte AND $01;     { either 1 or 0 }
                myByte := myByte SHR 1;
                Dec( myPlace )
            end
        end
end;


procedure TMarquee.DrawNormal;

{   Just display the new text.
}

var
    row,
    col:    Integer;

begin
    for col := 0 to FHorizontalPixels - 1 do
        for row := 0 to PIXEL_HEIGHT do
            FDisplayRack[ row, col ] := FWorkRack[ row, col ];

    Redraw
end;


procedure TMarquee.DrawRunLeft;

{   old string exits to the right then new string enters from the right
}

begin
end;


procedure TMarquee.DrawRunRight;

{   old string exits to the left then new string enters from the left.
}

begin
end;


procedure TMarquee.DrawChaseLeft;

{   old string exits to the left while the new string enters from the right
}

var
    row,
    col,
    bit:    Integer;

begin
    for col := 0 to maxCharacters - 1 do
    begin
        for row := 0 to PIXEL_HEIGHT do
        begin
            { shift all characters left by one position }
            for bit := 0 to FHorizontalPixels - 7 do
                FDisplayRack[ row, bit ] := FDisplayRack[ row, bit + 6 ];

            { place next character in rightmost position }
            for bit := FHorizontalPixels - 6 to FHorizontalPixels - 1 do
                FDisplayRack[ row, bit ] := FWorkRack[ row, ( col * 6 ) + ( bit - FHorizontalPixels - 6 ) ]
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawChaseRight;

{   old string exits to the right while the new string enters from the left
}

var
    row,
    col,
    bit:    Integer;

begin
    for col := 0 to maxCharacters - 1 do
    begin
        for row := 0 to PIXEL_HEIGHT do
        begin
            { shift all characters right by one position }
            for bit := FHorizontalPixels - 1 downto 6 do
                FDisplayRack[ row, bit ] := FDisplayRack[ row, bit - 6 ];

            { place next character (end characters first) in leftmost position }
            for bit := 0 to 5 do
                FDisplayRack[ row, bit ] := FWorkRack[ row, ( ( maxCharacters - col - 1 ) * 6 ) + bit ]
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawPopUp;

{   old string drops down then new string rises
}

var
    row,
    col,
    bit:    Integer;

begin

    { wipe out the existing display - drops one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to FHorizontalPixels - 1 do
        begin
            for row := PIXEL_HEIGHT downto bit + 1 do
                FDisplayRack[ row, col ] := FDisplayRack[ row - 1, col ];

            FDisplayRack[ bit, col ] := 0
        end;

        AnimationPause;
        Redraw
    end;

    { fill in the new display - rises one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to FHorizontalPixels - 1 do
        begin
            for row := ( PIXEL_HEIGHT - bit ) to PIXEL_HEIGHT - 1 do
                FDisplayRack[ row, col ] := FDisplayRack[ row + 1, col ];

            { filling from the bottom, drop in top bits first }
            FDisplayRack[ PIXEL_HEIGHT, col ] := FWorkRack[ bit, col ]
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawPopDown;

{   old string rises up then new string drops in
}

var
    row,
    col,
    bit:    Integer;

begin

    { wipe out the existing display - rises one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to FHorizontalPixels - 1 do
        begin
            for row := 0 to ( ( PIXEL_HEIGHT - 1 ) - bit ) do
                FDisplayRack[ row, col ] := FDisplayRack[ row + 1, col ];

            FDisplayRack[ PIXEL_HEIGHT - bit, col ] := 0
        end;

        AnimationPause;
        Redraw
    end;

    { fill in the new display - drops in one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to FHorizontalPixels - 1 do
        begin
            for row := bit downto 1 do
                FDisplayRack[ row, col ] := FDisplayRack[ row - 1, col ];

            { 7 - bit -- filling from the top, drop in bottom bits first }
            FDisplayRack[ 0, col ] := FWorkRack[ PIXEL_HEIGHT - bit, col ]
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawNewWhite;

{   Draws the characters one at a time, but the new character always appears in
    white.
}

var
    saveBrightColour:   TColor;
    
    bit,
    row,
    priorCol,
    col:                Integer;
    
begin
    { start by clearing the display }

    for row := 0 to PIXEL_HEIGHT do    
        for col := 0 to FHorizontalPixels - 1 do
            FDisplayRack[ row, col ] := 0;

    Redraw;

    { draw the characters, one at a time. Note one additional iteration to reset the last character }

    for col := 0 to maxCharacters do
    begin
        { first draw of the character; draw it in white }
        saveBrightColour := FBrightColour;
        FBrightColour := clWhite;

        if ( col < maxCharacters ) then        
            for row := 0 to PIXEL_HEIGHT do
                for bit := 0 to 5 do            
                    FDisplayRack[ row, ( col * 6 ) + bit ] := FWorkRack[ row, ( col * 6 ) + bit ];
                    
        Redraw;

        { draw the prior character; in the original color }
        FBrightColour := saveBrightColour;
        priorCol := col - 1;
        if ( col > 0 ) then
        begin
            priorCol := col - 1;
            for row := 0 to PIXEL_HEIGHT do
                for bit := 0 to 5 do  
                begin          
                    FPriorRack[ row, ( priorCol * 6 ) + bit ] := 0;     { so it gets regenerated! }
                    FDisplayRack[ row, ( priorCol * 6 ) + bit ] := FWorkRack[ row, ( priorCol * 6 ) + bit ]
                end
        end;                    
                           
        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawWipe;

{   Another cool way of displaying bits -- wipes individual characters
    then wipes the new characters in.
}

var
    row,
    col,
    bit:    Integer;

begin
    { wipe out the existing display }
    for bit := 0 to 5 do
    begin
        for col := 0 to maxCharacters - 1 do
            for row := 0 to PIXEL_HEIGHT do
                FDisplayRack[ row, ( col * 6 ) + bit ] := 0;

        AnimationPause;
        Redraw
    end;

    { wipe in the new display }

    for bit := 0 to 5 do
    begin
        for col := 0 to maxCharacters - 1 do
            for row := 0 to PIXEL_HEIGHT do
                FDisplayRack[ row, ( col * 6 ) + bit ] := FWorkRack[ row, ( col * 6 ) + bit ];

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawPopUpNew;

{   rotates upward only characters that have changed since the last setMessage().
    nice for odometer-style displays. See also DrawPopDownNew().
}

var
    row,
    col,
    bit,
    myLen,
    x,
    i:          Integer;
    changed:    array[ 0..MAX_CHARS ] of Boolean;

begin
    for i := 0 to maxCharacters - 1 do
        changed[ i ] := FALSE;

    myLen := Length( thisText );
    if ( myLen <> Length( prevText ) ) then
        for i := 1 to myLen do
            changed[ i - 1 ] := TRUE
    else
        for i := 1 to myLen do
            changed[ i - 1 ] := thisText[ i ] <> prevText[ i ];

    { rotate out the existing display - rises one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to maxCharacters - 1 do
            if ( changed[ col ] ) then    { rotate this character in }
                for i := 0 to 5 do
                begin
                    x := i + ( col * 6 );
                    for row := 0 to PIXEL_HEIGHT - 1 do
                        FDisplayRack[ row, x ] := FDisplayRack[ row + 1, x ];

                    FDisplayRack[ PIXEL_HEIGHT, x ] := FWorkRack[ bit, x ]
                end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawPopDownNew;

{   rotates downward only characters that have changed since the last setMessage().
    nice for odometer-style displays. See also DrawPopUpNew().
}

var 
    row,
    col,
    bit,
    myLen,
    x,
    i:          Integer;
    changed:    array[ 0..MAX_CHARS ] of Boolean;

begin
    for i := 0 to maxCharacters - 1 do
        changed[ i ] := FALSE;

    myLen := Length( thisText );

    { if the length of the string is different, all characters are changed. }
    if ( myLen <> Length( prevText ) ) then 
        for i := 1 to myLen do
            changed[ i - 1 ] := TRUE
    else
        for i := 1 to myLen do
            changed[ i - 1 ] := thisText[ i ] <> prevText[ i ];

    { rotate out the existing display - drops one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do    { eight draws }
    begin
        for col := 0 to maxCharacters - 1 do  { count characters }
            if ( changed[ col ] ) then    { rotate this character in }
                for i := 0 to 5 do      { count pixels across current character }
                begin
                    x := i + ( col * 6 );

                    { rotate bottom seven pixels down one position }
                    for row := PIXEL_HEIGHT downto 1 do
                        FDisplayRack[ row, x ] := FDisplayRack[ row - 1, x ];

                    { rotate next pixel, from bottom to top, into the top position }
                    FDisplayRack[ 0, x ] := FWorkRack[ PIXEL_HEIGHT - bit, x ]
                end;

        AnimationPause;
        Redraw
    end
end;



procedure TMarquee.DrawSlideChase;

{   like DrawSlide() except the new text immediately starts in as the old leaves.
    This has a different look to it and also executes faster than DrawSlide().

    Note: this tends to look more confusing than cool.
}

var
    bit,
    row,
    col:  Integer;

begin
    for col := 0 to maxCharacters - 1 do
    begin
        for row := 0 to PIXEL_HEIGHT do
        begin
            { shift all characters left by one position }

            if ( odd( row ) ) then
            begin
                for bit := 0 to FHorizontalPixels - 7 do
                    FDisplayRack[ row, bit ] := FDisplayRack[ row, bit + 6 ];

                { place next character in rightmost position }
                for bit := FHorizontalPixels - 6 to FHorizontalPixels - 1 do
                    FDisplayRack[ row, bit ] := FWorkRack[ row, ( col * 6 ) + ( bit - ( FHorizontalPixels - 6 ) ) ]
            end
            else
            begin
                { shift all characters right by one position }
                for bit := FHorizontalPixels - 1 downto 6 do
                    FDisplayRack[ row, bit ] := FDisplayRack[ row, bit - 6 ];

                { place next character (end characters first) in leftmost position }
                for bit := 0 to 5 do
                    FDisplayRack[ row, bit ] := FWorkRack[ row, ( ( maxCharacters - col - 1 ) * 6 ) + bit ]
            end
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawVerticalSlide;

{   Draws characters by shifting up/down characters by alternate bits (1,3,5 up
    and 2,4,6 down) or something like that.
}

var
    row,
    col,
    bit:    Integer;

begin

    { wipe out the existing display - drops one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to FHorizontalPixels - 1 do
        begin
            if ( odd( col ) ) then
            begin
                for row := PIXEL_HEIGHT downto bit + 1 do
                    FDisplayRack[ row, col ] := FDisplayRack[ row - 1, col ];

                FDisplayRack[ bit, col ] := 0
            end
            else
            begin
                for row := 0 to ( ( PIXEL_HEIGHT - 1 ) - bit ) do
                    FDisplayRack[ row, col ] := FDisplayRack[ row + 1, col ];

                FDisplayRack[ PIXEL_HEIGHT - bit, col ] := 0
            end
        end;

        AnimationPause;
        Redraw
    end;

    { fill in the new display - rises one bit at a time }

    for bit := 0 to PIXEL_HEIGHT do
    begin
        for col := 0 to FHorizontalPixels - 1 do
        begin
            if ( odd( col ) ) then
            begin
                for row := ( PIXEL_HEIGHT - bit ) to ( PIXEL_HEIGHT - 1 ) do
                    FDisplayRack[ row, col ] := FDisplayRack[ row + 1, col ];

                { filling from the bottom, drop in top bits first }
                FDisplayRack[ PIXEL_HEIGHT, col ] := FWorkRack[ bit, col ]
            end
            else
            begin
                for row := bit downto 1 do
                    FDisplayRack[ row, col ] := FDisplayRack[ row - 1, col ];

                { 7 - bit -- filling from the top, drop in bottom bits first }
                FDisplayRack[ 0, col ] := FWorkRack[ PIXEL_HEIGHT - bit, col ]
            end
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.DrawSlide;

{   Cool way of displaying bits by shifting out the existing bits and
    shifting in the new bits.
}


var
    bit,
    row,
    col:  Integer;

begin
    { shift out the existing display }
    for col := 0 to maxCharacters - 1 do    { one iteration per display character }
    begin
        for row := 0 to PIXEL_HEIGHT do
        begin
            if ( odd( row ) ) then  { shift odd rows left }
            begin
                for bit := 0 to FHorizontalPixels - ( ( col + 1 ) * 6 ) - 1 do
                    FDisplayRack[ row, bit ] := FDisplayRack[ row, bit + 6 ];

                { fill in last painted column with space }
                for bit := 0 to 5 do
                    FDisplayRack[ row, FHorizontalPixels - ( col * 6 ) - 1 + bit ] := 0
            end
            else    { shift even rows right }
            begin
                for bit := FHorizontalPixels - 1 downto ( col + 1 ) * 6 do
                    FDisplayRack[ row, bit ] := FDisplayRack[ row, bit - 6 ];

                for bit := 0 to 5 do
                    FDisplayRack[ row, bit + ( col * 6 ) ] := 0
            end
        end;

        AnimationPause;
        Redraw
    end;

    { the display is now blank. Shift in the new racks }

    for col := 0 to maxCharacters - 1 do
    begin
        for row := 0 to PIXEL_HEIGHT do
        begin
            if ( odd( row ) ) then  { shift odd rows in right }
            begin
                for bit := ( col * 6 ) + 5 downto 6 do
                    FDisplayRack[ row, bit ] := FDisplayRack[ row, bit - 6 ];

                { copy next bit in from FWorkRack (work end to front) to leftmost bit of display rack }
                for bit := 0 to 5 do
                    FDisplayRack[ row, bit ] := FWorkRack[ row, FHorizontalPixels - 6 - ( col * 6 ) + bit ]
            end
            else                    { shift even rows in left }
            begin
                for bit := FHorizontalPixels - ( col + 1 ) * 6  to FHorizontalPixels - 7 do
                    FDisplayRack[ row, bit ] := FDisplayRack[ row, bit + 6 ];

                { copy next bit in from FWorkRack (front to end) to rightmost bit of display rack }
                for bit := 0 to 5 do
                    FDisplayRack[ row, FHorizontalPixels + bit - 6 ] := FWorkRack[ row, bit + ( col * 6 ) ]
            end
        end;

        AnimationPause;
        Redraw
    end
end;


procedure TMarquee.SetProgress( myProgress: Longint );

{   Sets, in progress meter mode, the progress (out of maximum)
}

begin
    if ( prevProgress = myProgress ) then   { absolute has not changed }
        Exit;

    AssembleProgressBits( myProgress );

    Regen;
    prevProgress := myProgress
end;


procedure TMarquee.Regen;
var
    myDrawMode:         TDrawMode;

begin
    if ( ( not isAlive ) OR ( not Visible ) ) then
        Exit;

    if ( drawMode = dmRandom ) then
        myDrawMode := TDrawMode( Random( 14 ) + 1 )
    else
        myDrawMode := drawMode;

    case ( myDrawMode ) of
        dmNormal:           DrawNormal;
        dmSlide:            DrawSlide;
        dmVerticalSlide:    DrawVerticalSlide;
        dmSlideChase:       DrawSlideChase;
        dmWipe:             DrawWipe;
        dmNewWhite:         DrawNewWhite;
        dmRunLeft:          DrawRunLeft;
        dmRunRight:         DrawRunRight;
        dmChaseLeft:        DrawChaseLeft;
        dmChaseRight:       DrawChaseRight;
        dmPopUp:            DrawPopUp;
        dmPopDown:          DrawPopDown;
        dmPopUpNew:         DrawPopUpNew;
        dmPopDownNew:       DrawPopDownNew;
        dmPaintRandom:      DrawPaintRandom;
    end
end;

procedure TMarquee.SetMessage( const s: String32 );

{   Displays a center-justified message on the marquis.
}

var
    i,
    myJustification:    Integer;
    myDrawMode:         TDrawMode;
    myText:		        String32;

begin
    RecomputeMaxchars;

    { center the string then assemble the bit-pattern }
    captionText := s;

    case ( myAlignment ) of
        taLeftJustify:
        myText := Format( '%.*s', [ maxCharacters, captionText ] );

        taRightJustify:
        begin
            myJustification := ( maxCharacters - Length( s ) );

            if ( myJustification > 0 ) then
                myText := Format( '%*s%s', [ myJustification, ' ', s ] )
            else
                myText := Format( '%.*s', [ maxCharacters, s ] )
        end;

        taCenter:
        begin
            myJustification := ( maxCharacters - Length( s ) ) div 2;

            if ( myJustification > 0 ) then
                myText := Format( '%*s%s', [ myJustification, ' ', s ] )
            else
                myText := Format( '%.*s', [ maxCharacters, s ] )
        end
    end;


    thisText := myText;

    if ( not drawSame ) then                { are we skipping over if the text does not change? }
        if ( thisText = prevText ) then     { did the text not change? }
            Exit;                           { then forget the whole thing! }

    AssembleBits( myText );
    Regen;

    prevText := myText
end;


procedure TMarquee.Redraw;

{   Repaints the display. Only repaints pixels that have changed since the
    last paint (FDisplayRack vs. from FPriorRack). This assumes that either
    less than half the pixels repaint themselves each time or that assigning
    and comparing an integer is a lot cheaper (CPU-wise) than drawing four
    pixels.
}

var
    thisBit:    Byte;

    row,
    col:        Integer;

begin
    if ( ( not isAlive ) OR ( not Visible ) OR ( not bitPatternBuilt ) ) then
        Exit;

    if ( invalidateBrightOnly ) then    { special case of changing FBrightColour only }
    begin
        invalidateBrightOnly := FALSE;
        for col := 0 to FHorizontalPixels - 1 do
            for row := 0 to PIXEL_HEIGHT do
            begin
                thisBit := FDisplayRack[ row, col ];

                FPriorRack[ row, col ] := thisBit;
                if ( thisBit <> 0 ) then
                    DrawPixel( FLeftIndent + col * FPixelSize, FTopIndent + row * FPixelSize, FALSE )
            end
    end
    else if ( FInvalidateDisplay ) then
    begin
        FillRectangle;
        FInvalidateDisplay := FALSE;
        for col := 0 to FHorizontalPixels - 1 do
            for row := 0 to PIXEL_HEIGHT do
            begin
                thisBit := FDisplayRack[ row, col ];

                FPriorRack[ row, col ] := thisBit;
                DrawPixel( FLeftIndent + col * FPixelSize, FTopIndent + row * FPixelSize, thisBit = 0 )
            end
    end
    else
        for col := 0 to FHorizontalPixels - 1 do
            for row := 0 to PIXEL_HEIGHT do
            begin
                thisBit := FDisplayRack[ row, col ];
                if ( thisBit <> FPriorRack[ row, col ] ) then
                begin
                    FPriorRack[ row, col ] := thisBit;
                    DrawPixel( FLeftIndent + col * FPixelSize, FTopIndent + row * FPixelSize, thisBit = 0 )
                end
            end
end;


procedure TMarquee.Paint;
begin
    inherited Paint;

    RepaintMe
end;


procedure TMarquee.FillRectangle;
begin
    canvas.Brush.Color := Color;

    canvas.FillRect( Rect( 0, 0, Width, Height ) )
end;


procedure TMarquee.RepaintMe;

{   Repaints the entire display. Attached to an OnPaint event.
}

var
    row,
    col:        Integer;

begin
    if ( ( not isAlive ) OR ( not Visible ) ) then
        Exit;

    RecomputeMaxchars;
    FillRectangle;

    for col := 0 to FHorizontalPixels - 1 do
        for row := 0 to PIXEL_HEIGHT do
            DrawPixel( FLeftIndent + col * FPixelSize, FTopIndent + row * FPixelSize, FDisplayRack[ row, col ] = 0 )
end;


procedure TMarquee.DrawPixel( col, row: Integer; isDim: Boolean );

{   Draws a single zot on screen.
}

begin
    with ( Canvas ) do
    begin
        if ( FPixelSize = 2 ) then  { problems drawing a 1 x 1 rectangle. }
        begin
            if ( isDim ) then
                Pixels[ col, row ] := FDimColour
            else
                Pixels[ col, row ] := FBrightColour
        end
        else
        begin
            if ( isDim ) then
            begin
                Brush.Color := FDimColour;
                Pen.Color := FDimColour
            end
            else
            begin
                Brush.Color := FBrightColour;
                Pen.Color := FBrightColour
            end;
            
            Rectangle( col, row, col + FPixelSizeM1, row + FPixelSizeM1 )
        end
    end
end;


procedure TMarquee.SetPixelSize( pixels: Integer );

{   sets the size of a pixel; default is three.
}

begin
    FPixelSize := pixels;
    FPixelSizeM1 := pixels - 1;
    Canvas.Pen.Width := 1;
    
    RecomputeMaxchars;

    Regen
end;


procedure TMarquee.SetBrightColour( bright: TColor );

{   sets the colour of the "on" pixels

    inputs:
    bright  colour of "on" pixels

    outputs:
    none

    returns:
    none
}

begin
    FBrightColour := bright;
    invalidateBrightOnly := TRUE;
    DrawNormal
end;


procedure TMarquee.SetDimColour( dim: TColor );

{   sets the colour of the "off" pixels

    inputs:
    dim     colour of "off" pixels

    outputs:
    none

    returns:
    none
}

begin
    FDimColour := dim;
    FInvalidateDisplay := TRUE;
    DrawNormal
end;


procedure TMarquee.SetAlignment( how: TAlignment );
begin
    myAlignment := how;
    SetMessage( captionText )
end;

procedure TMarquee.AnimationPause;

{   pauses between redraws for animation effect.
}

var
    start:          Longint;
    milliseconds:   Integer;

begin
    case ( myAnimationSpeed ) of
        anSlow:     milliseconds := 150;
        anMedium:   milliseconds := 40;
        anFast:     Exit
    end;


    if ( delayInstance ) then
        Exit;   { prevent re-entrancy - these things have a way of "stack" ing up! }

    delayInstance := TRUE;
    start := GetTickCount;

    while ( ( GetTickCount < start + milliseconds ) AND ( GetTickCount >= start ) ) do
        Application.ProcessMessages;

    delayInstance := FALSE
end;


procedure TMarquee.SetColour( dim, bright: TColor );

{   sets the colour of the dim and bright pixels.

    inputs:
    dim     colour of "off" pixels
    bright  colour of "on" pixels

    outputs:
    none

    returns:
    none
}

begin
    FDimColour := dim;
    FBrightColour := bright;

    { invalidate the previous display forcing a complete redraw next time }
    FInvalidateDisplay := TRUE;
    DrawNormal
end;

end.















